import abc
import datetime
import errno
import itertools
from collections import namedtuple
import logging
import tempfile
import os
import shutil
import threading

from qiniu.compat import json, b as to_bytes, is_windows, is_linux, is_macos
from qiniu.utils import io_md5, dt2ts

from .endpoint import Endpoint
from .region import Region, ServiceName
from .default_client import qn_http_client
from .middleware import RetryDomainsMiddleware
from .single_flight import SingleFlight


class RegionsProvider:
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __iter__(self):
        """
        Returns
        -------
        Generator[Region, None, None]
        """


class MutableRegionsProvider(RegionsProvider):
    @abc.abstractmethod
    def set_regions(self, regions):
        """
        Parameters
        ----------
        regions: list[Region]
        """


# --- serializers for QueryRegionsProvider ---

def _get_region_from_query(data, **kwargs):
    preferred_scheme = kwargs.get('preferred_scheme')
    if not preferred_scheme:
        preferred_scheme = 'http'

    domain_path_map = {
        k: (k.value, 'domains')
        for k in ServiceName
        if k not in [ServiceName.UP_ACC]
    }
    domain_path_map[ServiceName.UP_ACC] = ('up', 'acc_domains')

    services = {
        # sn service name, dsn data service name
        sn: [
            Endpoint(h, default_scheme=preferred_scheme)
            for h in data.get(dsn, {}).get(k, [])
        ]
        for sn, (dsn, k) in domain_path_map.items()
    }

    return Region(
        region_id=data.get('region'),
        s3_region_id=data.get('s3', {}).get('region_alias', None),
        services=services,
        ttl=data.get('ttl', None)
    )


_query_regions_single_flight = SingleFlight()


class QueryRegionsProvider(RegionsProvider):
    def __init__(
        self,
        access_key,
        bucket_name,
        endpoints_provider,
        preferred_scheme='http',
        max_retry_times_per_endpoint=1,
    ):
        """
        Parameters
        ----------
        access_key: str
        bucket_name: str
        endpoints_provider: Iterable[Endpoint]
        preferred_scheme: str
        max_retry_times_per_endpoint: int
        """
        self.access_key = access_key
        self.bucket_name = bucket_name
        self.endpoints_provider = endpoints_provider
        self.preferred_scheme = preferred_scheme
        self.max_retry_times_per_endpoint = max_retry_times_per_endpoint

    def __iter__(self):
        endpoints_md5 = io_md5([
            to_bytes(e.host) for e in self.endpoints_provider
        ])
        flight_key = ':'.join([
            endpoints_md5,
            self.access_key,
            self.bucket_name
        ])
        regions = _query_regions_single_flight.do(flight_key, self.__fetch_regions)
        # change to `yield from` when min version of python update to >= 3.3
        for r in regions:
            yield r

    def __fetch_regions(self):
        endpoints = list(self.endpoints_provider)
        if not endpoints:
            raise ValueError('There aren\'t any available endpoints to query regions')
        endpoint, alternative_endpoints = endpoints[0], endpoints[1:]

        url = '{0}/v4/query?ak={1}&bucket={2}'.format(endpoint.get_value(), self.access_key, self.bucket_name)
        ret, resp = qn_http_client.get(
            url,
            middlewares=[
                RetryDomainsMiddleware(
                    backup_domains=[e.host for e in alternative_endpoints],
                    max_retry_times=self.max_retry_times_per_endpoint
                )
            ]
        )

        if not resp.ok():
            raise RuntimeError(
                (
                    'Query regions failed with '
                    'HTTP Status Code {0}, '
                    'Body {1}'
                ).format(resp.status_code, resp.text_body)
            )

        return [
            _get_region_from_query(d, preferred_scheme=self.preferred_scheme)
            for d in ret.get('hosts', [])
        ]


# --- helpers for CachedRegionsProvider ---
class FileAlreadyLocked(RuntimeError):
    def __init__(self, message):
        super(FileAlreadyLocked, self).__init__(message)


_file_threading_lockers_lock = threading.Lock()
_file_threading_lockers = {}


class _FileThreadingLocker:
    def __init__(self, fd):
        self._fd = fd

    def __enter__(self):
        with _file_threading_lockers_lock:
            threading_lock = _file_threading_lockers.get(self._file_path, threading.Lock())
            # Could use keyword style `acquire(blocking=False)` when min version of python update to >= 3
            if not threading_lock.acquire(False):
                raise FileAlreadyLocked('File {0} already locked'.format(self._file_path))
            _file_threading_lockers[self._file_path] = threading_lock

    def __exit__(self, exc_type, exc_val, exc_tb):
        with _file_threading_lockers_lock:
            threading_lock = _file_threading_lockers.get(self._file_path)
            if threading_lock and threading_lock.locked():
                threading_lock.release()
                del _file_threading_lockers[self._file_path]

    @property
    def _file_path(self):
        return self._fd.name


if is_linux or is_macos:
    import fcntl

    # Use subclass of _FileThreadingLocker when min version of python update to >= 3
    class _FileLocker:
        def __init__(self, fd):
            self._fd = fd

        def __enter__(self):
            try:
                fcntl.lockf(self._fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except IOError:
                # Use `raise ... from ...` when min version of python update to >= 3
                raise FileAlreadyLocked('File {0} already locked'.format(self._file_path))

        def __exit__(self, exc_type, exc_val, exc_tb):
            fcntl.lockf(self._fd, fcntl.LOCK_UN)

        @property
        def _file_path(self):
            return self._fd.name

elif is_windows:
    import msvcrt

    class _FileLocker:
        def __init__(self, fd):
            self._fd = fd
            self._lock_fd = None
            self._already_locked = False

        def __enter__(self):
            try:
                self._lock_fd = open(self._lock_file_path, 'w')
                msvcrt.locking(self._lock_fd.fileno(), msvcrt.LK_LOCK | msvcrt.LK_NBLCK, 1)
            except OSError:
                self._already_locked = True
                raise FileAlreadyLocked('File {0} already locked'.format(self._file_path))

        def __exit__(self, exc_type, exc_val, exc_tb):
            if self._already_locked:
                if self._lock_fd:
                    self._lock_fd.close()
                return

            try:
                msvcrt.locking(self._lock_fd.fileno(), msvcrt.LK_UNLCK, 1)
            finally:
                self._lock_fd.close()
                os.remove(self._lock_file_path)

        @property
        def _file_path(self):
            return self._fd.name

        @property
        def _lock_file_path(self):
            """
            Returns
            -------
            str
            """
            return self._file_path + '.lock'

else:
    class _FileLocker:
        def __init__(self, fd):
            self._fd = fd
            self._already_locked = False

        def __enter__(self):
            try:
                # Atomic file creation
                open_flags = os.O_EXCL | os.O_RDWR | os.O_CREAT
                fd = os.open(self._lock_file_path, open_flags)
                os.close(fd)
            except OSError:
                self._already_locked = True
                raise FileAlreadyLocked('File {0} already locked'.format(self._file_path))

        def __exit__(self, exc_type, exc_val, exc_tb):
            if self._already_locked:
                return
            try:
                os.remove(self._lock_file_path)
            except OSError:
                pass

        @property
        def _file_path(self):
            return self._fd.name

        @property
        def _lock_file_path(self):
            """
            Returns
            -------
            str
            """
            return self._file_path + '.lock'


# use dataclass instead namedtuple if min version of python update to 3.7
CacheScope = namedtuple(
    'CacheScope',
    [
        'memo_cache',
        'persist_path',
        'last_shrink_at',
        'shrink_interval',
        'should_shrink_expired_regions',
        'memo_cache_lock'
    ]
)


_global_cache_scope = CacheScope(
    memo_cache={},
    persist_path=os.path.join(
        tempfile.gettempdir(),
        'qn-py-sdk',
        'regions-cache.jsonl'
    ),
    last_shrink_at=datetime.datetime.fromtimestamp(0),
    shrink_interval=datetime.timedelta(days=1),
    should_shrink_expired_regions=False,
    memo_cache_lock=threading.Lock()
)


# --- serializers for CachedRegionsProvider ---

_PersistedEndpoint = namedtuple(
    'PersistedEndpoint',
    [
        'host',
        'defaultScheme'
    ]
)


def _persist_endpoint(endpoint):
    """
    Parameters
    ----------
    endpoint: Endpoint

    Returns
    -------
    dict
    """
    return _PersistedEndpoint(
        defaultScheme=endpoint.default_scheme,
        host=endpoint.host
    )._asdict()


def _get_endpoint_from_persisted(data):
    """
    Parameters
    ----------
    data: dict

    Returns
    -------
    Endpoint
    """
    persisted_endpoint = _PersistedEndpoint(**data)
    return Endpoint(
        persisted_endpoint.host,
        default_scheme=persisted_endpoint.defaultScheme
    )


_PersistedRegion = namedtuple(
    'PersistedRegion',
    [
        'regionId',
        's3RegionId',
        'services',
        'ttl',
        'createTime'
    ]
)


def _persist_region(region):
    """
    Parameters
    ----------
    region: Region

    Returns
    -------
    dict
    """
    return _PersistedRegion(
        regionId=region.region_id,
        s3RegionId=region.s3_region_id,
        services={
            # The StrEnum not available in python < 3.11
            # so need stringify the key manually
            k.value if isinstance(k, ServiceName) else k: [
                _persist_endpoint(e)
                for e in v
            ]
            for k, v in region.services.items()
        },
        ttl=region.ttl,
        # use datetime.datetime.timestamp() when min version of python >= 3
        createTime=dt2ts(region.create_time)
    )._asdict()


def _get_region_from_persisted(data):
    """
    Parameters
    ----------
    data: dict

    Returns
    -------
    Region
    """
    def _get_service_name(k):
        try:
            return ServiceName(k)
        except ValueError:
            return k

    persisted_region = _PersistedRegion(**data)

    return Region(
        region_id=persisted_region.regionId,
        s3_region_id=persisted_region.s3RegionId,
        services={
            # The StrEnum not available in python < 3.11
            # so need parse the key manually
            _get_service_name(k): [
                _get_endpoint_from_persisted(d)
                for d in v
            ]
            for k, v in persisted_region.services.items()
        },
        ttl=persisted_region.ttl,
        create_time=datetime.datetime.fromtimestamp(persisted_region.createTime / 1000)
    )


def _parse_persisted_regions(persisted_data):
    """
    Parameters
    ----------
    persisted_data: str

    Returns
    -------
    cache_key: str
    regions: list[Region]
    """
    parsed_data = json.loads(persisted_data)
    regions = [
        _get_region_from_persisted(d)
        for d in parsed_data.get('regions', [])
    ]
    return parsed_data.get('cacheKey'), regions


def _walk_persist_cache_file(persist_path, ignore_parse_error=True):
    """
    Parameters
    ----------
    persist_path: str
    ignore_parse_error: bool

    Returns
    -------
    Iterable[(str, list[Region])]
    """
    if not os.access(persist_path, os.R_OK):
        return

    with open(persist_path, 'r') as f:
        for line in f:
            if not line.strip():
                continue
            try:
                cache_key, regions = _parse_persisted_regions(line.strip())
                yield cache_key, regions
            except Exception as err:
                if not ignore_parse_error:
                    raise err


def _merge_regions(*args):
    """
    merge two regions by region id.
    if the same region id, the last create region will be keep.
    Parameters
    ----------
    args: list[Region]

    Returns
    -------
    list[Region]
    """
    regions_dict = {}

    for r in itertools.chain(*args):
        if r.region_id not in regions_dict:
            regions_dict[r.region_id] = r
        else:
            if r.create_time > regions_dict[r.region_id].create_time:
                regions_dict[r.region_id] = r

    return regions_dict.values()


class CachedRegionsProvider(MutableRegionsProvider):
    def __init__(
        self,
        cache_key,
        base_regions_provider,
        **kwargs
    ):
        """
        Parameters
        ----------
        cache_key: str
        base_regions_provider: Iterable[Region]
        kwargs
            persist_path: str
            shrink_interval: datetime.timedelta
            should_shrink_expired_regions: bool
        """
        self.cache_key = cache_key
        self.base_regions_provider = base_regions_provider

        persist_path = kwargs.get('persist_path', None)
        last_shrink_at = datetime.datetime.fromtimestamp(0)
        if persist_path is None:
            cache_dir = os.path.dirname(_global_cache_scope.persist_path)
            try:
                # make sure the cache dir is available for all users.
                # we can not use the '/tmp' dir directly on linux,
                # because the permission is 0o1777
                if not os.path.exists(cache_dir):
                    # os.makedirs have no exists_ok parameter in python 2.7
                    os.makedirs(cache_dir)
                    os.chmod(cache_dir, 0o777)
                persist_path = _global_cache_scope.persist_path
                last_shrink_at = _global_cache_scope.last_shrink_at
            except Exception as err:
                if isinstance(err, OSError) and err.errno == errno.EEXIST:
                    persist_path = _global_cache_scope.persist_path
                    last_shrink_at = _global_cache_scope.last_shrink_at
                else:
                    logging.warning(
                        'failed to create cache dir %s. error: %s', cache_dir, err)

        shrink_interval = kwargs.get('shrink_interval', None)
        if shrink_interval is None:
            shrink_interval = _global_cache_scope.shrink_interval

        should_shrink_expired_regions = kwargs.get('should_shrink_expired_regions', None)
        if should_shrink_expired_regions is None:
            should_shrink_expired_regions = _global_cache_scope.should_shrink_expired_regions

        self._cache_scope = _global_cache_scope._replace(
            persist_path=persist_path,
            last_shrink_at=last_shrink_at,
            shrink_interval=shrink_interval,
            should_shrink_expired_regions=should_shrink_expired_regions
        )

    def __iter__(self):
        if self.__should_shrink:
            try:
                self.__shrink_cache()
            except Exception as err:
                logging.warning('failed to shrink cache. error: %s', err)

        get_regions_fns = [
            self.__get_regions_from_memo,
            self.__get_regions_from_file,
            self.__get_regions_from_base_provider
        ]

        # set the fallback to None for raise errors when failed
        regions = None
        for get_regions in get_regions_fns:
            regions = get_regions(fallback=regions)
            if regions and all(r.is_live for r in regions):
                break

        # change to `yield from` when min version of python update to >= 3.3
        for r in regions:
            yield r

    def set_regions(self, regions):
        """
        Parameters
        ----------
        regions: list[Region]
        """
        with self._cache_scope.memo_cache_lock:
            self._cache_scope.memo_cache[self.cache_key] = regions

        if not self._cache_scope.persist_path:
            return

        try:
            with open(self._cache_scope.persist_path, 'a') as f:
                f.write(json.dumps({
                    'cacheKey': self.cache_key,
                    'regions': [_persist_region(r) for r in regions]
                }) + os.linesep)
        except Exception as err:
            logging.warning('failed to cache regions result to file. error: %s', err)

    @property
    def persist_path(self):
        """
        Returns
        -------
        str
        """
        return self._cache_scope.persist_path

    @persist_path.setter
    def persist_path(self, value):
        """
        Parameters
        ----------
        value: str
        """
        if value == self._cache_scope.persist_path:
            return
        self._cache_scope = self._cache_scope._replace(
            persist_path=value,
            last_shrink_at=datetime.datetime.fromtimestamp(0)
        )

    @property
    def last_shrink_at(self):
        """
        Returns
        -------
        datetime.datetime
        """
        # copy the datetime make sure it is read-only
        return self._cache_scope.last_shrink_at.replace()

    @property
    def shrink_interval(self):
        """
        Returns
        -------
        datetime.timedelta
        """
        return self._cache_scope.shrink_interval

    @shrink_interval.setter
    def shrink_interval(self, value):
        """
        Parameters
        ----------
        value: datetime.timedelta
        """
        self._cache_scope = self._cache_scope._replace(
            shrink_interval=value
        )

    @property
    def should_shrink_expired_regions(self):
        """
        Returns
        -------
        bool
        """
        return self._cache_scope.should_shrink_expired_regions

    @should_shrink_expired_regions.setter
    def should_shrink_expired_regions(self, value):
        """
        Parameters
        ----------
        value: bool
        """
        self._cache_scope = self._cache_scope._replace(
            should_shrink_expired_regions=value
        )

    def __get_regions_from_memo(self, fallback=None):
        """
        Parameters
        ----------
        fallback: list[Region]

        Returns
        -------
        list[Region]
        """
        regions = self._cache_scope.memo_cache.get(self.cache_key)

        if regions:
            return regions

        return fallback

    def __get_regions_from_file(self, fallback=None):
        """
        Parameters
        ----------
        fallback: list[Region]

        Returns
        -------
        list[Region]
        """
        if not self._cache_scope.persist_path:
            return fallback

        try:
            self.__flush_file_cache_to_memo()
        except Exception as err:
            if fallback is not None:
                return fallback
            else:
                raise err

        return self.__get_regions_from_memo(fallback)

    def __get_regions_from_base_provider(self, fallback=None):
        """
        Parameters
        ----------
        fallback: list[Region]

        Returns
        -------
        list[Region]
        """
        try:
            regions = list(self.base_regions_provider)
        except Exception as err:
            if fallback is not None:
                return fallback
            else:
                raise err
        self.set_regions(regions)
        return regions

    def __flush_file_cache_to_memo(self):
        for cache_key, regions in _walk_persist_cache_file(
            persist_path=self._cache_scope.persist_path
        ):
            if cache_key not in self._cache_scope.memo_cache:
                self._cache_scope.memo_cache[cache_key] = regions
                return
            memo_regions = self._cache_scope.memo_cache[cache_key]
            self._cache_scope.memo_cache[cache_key] = _merge_regions(
                memo_regions,
                regions
            )

    @property
    def __should_shrink(self):
        """
        Returns
        -------
        bool
        """
        return datetime.datetime.now() - self._cache_scope.last_shrink_at > self._cache_scope.shrink_interval

    def __shrink_cache(self):
        # shrink memory cache
        if self._cache_scope.should_shrink_expired_regions:
            memo_cache_old = self._cache_scope.memo_cache.copy()
            # Could use keyword style `acquire(blocking=False)` when min version of python update to >= 3
            if self._cache_scope.memo_cache_lock.acquire(False):
                try:
                    for k, regions in memo_cache_old.items():
                        live_regions = [r for r in regions if r.is_live]
                        if live_regions:
                            self._cache_scope.memo_cache[k] = live_regions
                        else:
                            del self._cache_scope.memo_cache[k]
                finally:
                    self._cache_scope.memo_cache_lock.release()

        # shrink file cache
        if not self._cache_scope.persist_path:
            self._cache_scope = self._cache_scope._replace(
                last_shrink_at=datetime.datetime.now()
            )
            return

        shrink_file_path = self._cache_scope.persist_path + '.shrink'
        try:
            with open(shrink_file_path, 'a') as f, _FileThreadingLocker(f), _FileLocker(f):
                # filter data
                shrunk_cache = {}
                for cache_key, regions in _walk_persist_cache_file(
                    persist_path=self._cache_scope.persist_path
                ):
                    kept_regions = regions
                    if self._cache_scope.should_shrink_expired_regions:
                        kept_regions = [
                            r for r in kept_regions if r.is_live
                        ]

                    if cache_key not in shrunk_cache:
                        shrunk_cache[cache_key] = kept_regions
                    else:
                        shrunk_cache[cache_key] = _merge_regions(
                            shrunk_cache[cache_key],
                            kept_regions
                        )

                # write data
                for cache_key, regions in shrunk_cache.items():
                    f.write(
                        json.dumps(
                            {
                                'cacheKey': cache_key,
                                'regions': [_persist_region(r) for r in regions]
                            }
                        ) + os.linesep
                    )

                # make the cache file available for all users
                if is_linux or is_macos:
                    os.chmod(shrink_file_path, 0o666)

                # rename file
                if is_windows:
                    # windows must close first, or will raise permission error
                    # be careful to do something with the file after this
                    f.close()
                shutil.move(shrink_file_path, self._cache_scope.persist_path)

                # update last shrink time
                self._cache_scope = self._cache_scope._replace(
                    last_shrink_at=datetime.datetime.now()
                )
                global _global_cache_scope
                if _global_cache_scope.persist_path == self._cache_scope.persist_path:
                    _global_cache_scope = _global_cache_scope._replace(
                        last_shrink_at=self._cache_scope.last_shrink_at
                    )

        except FileAlreadyLocked:
            # skip file shrink by another running
            pass


def get_default_regions_provider(
    query_endpoints_provider,
    access_key,
    bucket_name,
    accelerate_uploading=False,
    force_query=False,
    **kwargs
):
    """
    Parameters
    ----------
    query_endpoints_provider: Iterable[Endpoint]
    access_key: str
    bucket_name: str
    accelerate_uploading: bool
    force_query: bool
    kwargs
        preferred_scheme: str
            option of QueryRegionsProvider
        max_retry_times_per_endpoint: int
            option of QueryRegionsProvider
        persist_path: str
            option of CachedRegionsProvider
        shrink_interval: datetime.timedelta
            option of CachedRegionsProvider
        should_shrink_expired_regions: bool
            option of CachedRegionsProvider

    Returns
    -------
    Iterable[Region]
    """
    query_regions_provider_opts = {
        'access_key': access_key,
        'bucket_name': bucket_name,
        'endpoints_provider': query_endpoints_provider,
    }
    query_regions_provider_opts.update({
        k: v
        for k, v in kwargs.items()
        if k in ['preferred_scheme', 'max_retry_times_per_endpoint']
    })

    query_regions_provider = QueryRegionsProvider(**query_regions_provider_opts)

    if force_query:
        return query_regions_provider

    query_endpoints = list(query_endpoints_provider)

    endpoints_md5 = io_md5([
        to_bytes(e.host) for e in query_endpoints
    ])
    cache_key = ':'.join([
        endpoints_md5,
        access_key,
        bucket_name,
        'true' if accelerate_uploading else 'false'
    ])

    cached_regions_provider_opts = {
        'cache_key': cache_key,
        'base_regions_provider': query_regions_provider,
    }
    cached_regions_provider_opts.update({
        k: v
        for k, v in kwargs.items()
        if k in [
            'persist_path',
            'shrink_interval',
            'should_shrink_expired_regions'
        ]
    })

    return CachedRegionsProvider(
        **cached_regions_provider_opts
    )



from .form_uploader import FormUploader
from .resume_uploader_v1 import ResumeUploaderV1
from .resume_uploader_v2 import ResumeUploaderV2

__all__ = [
    'FormUploader',
    'ResumeUploaderV1',
    'ResumeUploaderV2'
]



from datetime import datetime, timedelta

from enum import Enum

from .endpoint import Endpoint


# Use StrEnum when min version of python update to >= 3.11
# to make the json stringify more readable,
# or find another way to simple the json stringify
class ServiceName(Enum):
    UC = 'uc'
    UP = 'up'
    UP_ACC = 'up_acc'
    IO = 'io'
    # IO_SRC = 'io_src'
    RS = 'rs'
    RSF = 'rsf'
    API = 'api'


class Region:
    @staticmethod
    def merge(*args):
        """
        Parameters
        ----------
        args: list[list[Region]]

        Returns
        -------

        """
        if not args:
            raise TypeError('There aren\'ta any regions to merge')
        source, rest = args[0], args[1:]
        target = source.clone()
        for r in rest:
            for sn, el in r.services.items():
                if sn not in target.services:
                    target.services[sn] = [e.clone() for e in el]
                else:
                    target_values = [e.get_value() for e in target.services[sn]]
                    target.services[sn] += [
                        e.clone()
                        for e in el
                        if e.get_value() not in target_values
                    ]

        return target

    @staticmethod
    def from_region_id(region_id, **kwargs):
        """
        Parameters
        ----------
        region_id: str
        kwargs: dict
            s3_region_id: str
            ttl: int
            create_time: datetime
            extended_services: dict[str, list[Region]]
            preferred_scheme: str

        Returns
        -------
        Region
        """
        # create services endpoints
        endpoint_kwargs = {
        }
        if 'preferred_scheme' in kwargs:
            endpoint_kwargs['default_scheme'] = kwargs.get('preferred_scheme')

        is_z0 = region_id == 'z0'
        services_hosts = {
            ServiceName.UC: ['uc.qiniuapi.com'],
            ServiceName.UP: [
                'upload-{0}.qiniup.com'.format(region_id),
                'up-{0}.qiniup.com'.format(region_id)
            ] if not is_z0 else [
                'upload.qiniup.com',
                'up.qiniup.com'
            ],
            ServiceName.IO: [
                'iovip-{0}.qiniuio.com'.format(region_id),
            ] if not is_z0 else [
                'iovip.qiniuio.com',
            ],
            ServiceName.RS: [
                'rs-{0}.qiniuapi.com'.format(region_id),
            ],
            ServiceName.RSF: [
                'rsf-{0}.qiniuapi.com'.format(region_id),
            ],
            ServiceName.API: [
                'api-{0}.qiniuapi.com'.format(region_id),
            ]
        }
        services = {
            k: [
                Endpoint(h, **endpoint_kwargs) for h in v
            ]
            for k, v in services_hosts.items()
        }
        services.update(kwargs.get('extended_services', {}))

        # create region
        region_kwargs = {
            k: kwargs.get(k)
            for k in [
                's3_region_id',
                'ttl',
                'create_time'
            ] if k in kwargs
        }
        region_kwargs['region_id'] = region_id
        region_kwargs.setdefault('s3_region_id', region_id)
        region_kwargs['services'] = services

        return Region(**region_kwargs)

    def __init__(
        self,
        region_id=None,
        s3_region_id=None,
        services=None,
        ttl=86400,
        create_time=None
    ):
        """
        Parameters
        ----------
        region_id: str
        s3_region_id: str
        services: dict[ServiceName or str, list[Endpoint]]
        ttl: int, default 86400
        create_time: datetime, default datetime.now()
        """
        self.region_id = region_id
        self.s3_region_id = s3_region_id if s3_region_id else region_id

        self.services = services if services else {}
        self.services.update(
            {
                k: []
                for k in ServiceName
                if
                k not in self.services or
                not isinstance(self.services[k], list)
            }
        )

        self.ttl = ttl
        self.create_time = create_time if create_time else datetime.now()

    @property
    def is_live(self):
        """
        Returns
        -------
        bool
        """
        if self.ttl < 0:
            return True
        live_time = datetime.now() - self.create_time
        return live_time < timedelta(seconds=self.ttl)

    def clone(self):
        """
        Returns
        -------
        Region
        """
        return Region(
            region_id=self.region_id,
            s3_region_id=self.s3_region_id,
            services={
                k: [endpoint.clone() for endpoint in self.services[k]]
                for k in self.services
            },
            ttl=self.ttl,
            create_time=self.create_time
        )



from collections import namedtuple

from qiniu.http.endpoints_retry_policy import EndpointsRetryPolicy
from qiniu.http.region import ServiceName
from qiniu.http.regions_retry_policy import RegionsRetryPolicy
from qiniu.retry.abc import RetryPolicy
from qiniu.retry import Retrier


_TokenExpiredRetryState = namedtuple(
    'TokenExpiredRetryState',
    [
        'retried_times',
        'upload_api_version'
    ]
)


class TokenExpiredRetryPolicy(RetryPolicy):
    def __init__(
        self,
        upload_api_version,
        record_delete_handler,
        record_exists_handler,
        max_retry_times=1
    ):
        """
        Parameters
        ----------
        upload_api_version: str
        record_delete_handler: callable
            `() -> None`
        record_exists_handler: callable
            `() -> bool`
        max_retry_times: int
        """
        self.upload_api_version = upload_api_version
        self.record_delete_handler = record_delete_handler
        self.record_exists_handler = record_exists_handler
        self.max_retry_times = max_retry_times

    def init_context(self, context):
        """
        Parameters
        ----------
        context: dict
        """
        context[self] = _TokenExpiredRetryState(
            retried_times=0,
            upload_api_version=self.upload_api_version
        )

    def should_retry(self, attempt):
        """
        Parameters
        ----------
        attempt: qiniu.retry.Attempt

        Returns
        -------
        bool
        """
        state = attempt.context[self]

        if (
            state.retried_times >= self.max_retry_times or
            not self.record_exists_handler()
        ):
            return False

        if not attempt.result:
            return False

        _ret, resp = attempt.result

        if (
            state.upload_api_version == 'v1' and
            resp.status_code == 701
        ):
            return True

        if (
            state.upload_api_version == 'v2' and
            resp.status_code == 612
        ):
            return True

        return False

    def prepare_retry(self, attempt):
        """
        Parameters
        ----------
        attempt: qiniu.retry.Attempt
        """
        state = attempt.context[self]
        attempt.context[self] = state._replace(retried_times=state.retried_times + 1)

        if not self.record_exists_handler():
            return

        self.record_delete_handler()


class AccUnavailableRetryPolicy(RetryPolicy):
    def __init__(self):
        pass

    def init_context(self, context):
        pass

    def should_retry(self, attempt):
        """
        Parameters
        ----------
        attempt: qiniu.retry.Attempt

        Returns
        -------
        bool
        """
        if not attempt.result:
            return False

        region = attempt.context.get('region')
        if not region:
            return False

        if all(
            not region.services[sn]
            for sn in attempt.context.get('alternative_service_names')
        ):
            return False

        _ret, resp = attempt.result

        return resp.status_code == 400 and \
            'transfer acceleration is not configured on this bucket' in resp.text_body

    def prepare_retry(self, attempt):
        """
        Parameters
        ----------
        attempt: qiniu.retry.Attempt
        """
        endpoints = []
        while not endpoints:
            if not attempt.context.get('alternative_service_names'):
                raise RuntimeError('No alternative service available')
            attempt.context['service_name'] = attempt.context.get('alternative_service_names').pop(0)
            # shallow copy list
            # change to `list.copy` for more readable when min version of python update to >= 3
            endpoints = attempt.context['region'].services.get(attempt.context['service_name'], [])[:]
        attempt.context['alternative_endpoints'] = endpoints
        attempt.context['endpoint'] = attempt.context['alternative_endpoints'].pop(0)


ProgressRecord = namedtuple(
    'ProgressRecorder',
    [
        'upload_api_version',
        'exists',
        'delete'
    ]
)


def get_default_retrier(
    regions_provider,
    preferred_endpoints_provider=None,
    progress_record=None,
    accelerate_uploading=False
):
    """
    Parameters
    ----------
    regions_provider: Iterable[Region]
    preferred_endpoints_provider: Iterable[Endpoint]
    progress_record: ProgressRecord
    accelerate_uploading: bool

    Returns
    -------
    Retrier
    """
    retry_policies = []
    upload_service_names = [ServiceName.UP]
    handle_change_region = None

    if accelerate_uploading:
        retry_policies.append(AccUnavailableRetryPolicy())
        upload_service_names.insert(0, ServiceName.UP_ACC)

    if progress_record:
        retry_policies.append(TokenExpiredRetryPolicy(
            upload_api_version=progress_record.upload_api_version,
            record_delete_handler=progress_record.delete,
            record_exists_handler=progress_record.exists
        ))

        def _handle_change_region(_):
            progress_record.delete()

        handle_change_region = _handle_change_region

    retry_policies += [
        EndpointsRetryPolicy(skip_init_context=True),
        RegionsRetryPolicy(
            regions_provider=regions_provider,
            service_names=upload_service_names,
            preferred_endpoints_provider=preferred_endpoints_provider,
            on_change_region=handle_change_region
        )
    ]

    return Retrier(retry_policies)



# -*- coding: utf-8 -*-
import io
import os

from qiniu.compat import is_py2, str as text_type

from .client import SandboxClient


def load_dotenv_if_present(*paths):
    if not paths:
        paths = (
            os.path.join(os.getcwd(), '.env'),
        )

    for path in paths:
        if not path or not os.path.exists(path):
            continue
        with io.open(path, 'r', encoding='utf-8') as f:
            for raw_line in f:
                line = raw_line.strip()
                if not line or line.startswith('#') or '=' not in line:
                    continue
                key, value = line.split('=', 1)
                key = key.strip()
                value = _strip_inline_comment(value.strip()).strip()
                if (
                    len(value) >= 2 and
                    value[0] == value[-1] and
                    value[0] in ('"', "'")
                ):
                    value = value[1:-1]
                key, value = _native_env_pair(key, value)
                if key and key not in os.environ:
                    os.environ[key] = value


def _native_env_pair(key, value):
    if is_py2:
        if isinstance(key, text_type):
            key = key.encode('utf-8')
        if isinstance(value, text_type):
            value = value.encode('utf-8')
    return key, value


def _strip_inline_comment(value):
    quote = None
    escaped = False
    for index, char in enumerate(value):
        if escaped:
            escaped = False
            continue
        if char == '\\':
            escaped = True
            continue
        if quote:
            if char == quote:
                quote = None
            continue
        if char in ('"', "'"):
            quote = char
            continue
        if char == '#' and (index == 0 or value[index - 1].isspace()):
            return value[:index]
    return value


def env(key, fallback=None):
    return os.getenv(key) or fallback


def first_env(*keys):
    for key in keys:
        value = os.getenv(key)
        if value:
            return value
    return None


def required_env(key):
    value = os.getenv(key)
    if value:
        return value
    raise RuntimeError('Please set {0}'.format(key))


def sandbox_endpoint():
    return env('QINIU_SANDBOX_ENDPOINT')


def sandbox_api_key():
    value = first_env(
        'QINIU_SANDBOX_API_KEY',
        'QINIU_API_KEY',
        'E2B_API_KEY',
    )
    if value:
        return value
    raise RuntimeError(
        'Please set QINIU_SANDBOX_API_KEY, QINIU_API_KEY, or E2B_API_KEY')


def sandbox_template():
    return env('QINIU_SANDBOX_TEMPLATE', 'base')


def sandbox_client(**options):
    defaults = {
        'endpoint': sandbox_endpoint(),
        'api_key': sandbox_api_key(),
    }
    access_key = env('QINIU_SANDBOX_ACCESS_KEY')
    secret_key = env('QINIU_SANDBOX_SECRET_KEY')
    if access_key and secret_key:
        defaults['access_key'] = access_key
        defaults['secret_key'] = secret_key
    defaults.update(options)
    return SandboxClient(**defaults)



import abc


class RetryPolicy(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def init_context(self, context):
        """
        initial context values the policy required

        Parameters
        ----------
        context: dict
        """

    @abc.abstractmethod
    def should_retry(self, attempt):
        """
        if returns True, this policy will be applied

        Parameters
        ----------
        attempt: qiniu.retry.attempt.Attempt

        Returns
        -------
        bool
        """

    @abc.abstractmethod
    def prepare_retry(self, attempt):
        """
        apply this policy to change the context values for next attempt

        Parameters
        ----------
        attempt: qiniu.retry.attempt.Attempt
        """

    def is_important(self, attempt):
        """
        if returns True, this policy will be applied, whether it should retry or not.
        this is useful when want to stop retry.

        Parameters
        ----------
        attempt: qiniu.retry.attempt.Attempt

        Returns
        -------
        bool
        """

    def after_retry(self, attempt, policy):
        """
        Parameters
        ----------
        attempt: qiniu.retry.attempt.Attempt
        policy: RetryPolicy
        """



# -*- coding: utf-8 -*-
'''
Qiniu Resource Storage SDK for Python
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

For detailed document, please see:
<https://developer.qiniu.com/kodo/sdk/1242/python>
'''

# flake8: noqa

__version__ = '7.18.1'

from .auth import Auth, QiniuMacAuth

from .config import set_default
from .zone import Zone
from .region import LegacyRegion as Region

from .services.storage.bucket import BucketManager, build_batch_copy, build_batch_rename, build_batch_move, \
    build_batch_stat, build_batch_delete, build_batch_restoreAr, build_batch_restore_ar
from .services.storage.uploader import put_data, put_file, put_file_v2, put_stream, put_stream_v2
from .services.storage.upload_progress_recorder import UploadProgressRecorder
from .services.cdn.manager import CdnManager, DataType, create_timestamp_anti_leech_url, DomainManager
from .services.processing.pfop import PersistentFop
from .services.processing.cmd import build_op, pipe_cmd, op_save
from .services.compute.app import AccountClient
from .services.compute.qcos_api import QcosClient
from .services.sandbox import (
    EntryInfo,
    FileType,
    GitAuthException,
    GitBranches,
    GitFileStatus,
    FilesystemEventType,
    GitRepositoryResource,
    GitStatus,
    GitUpstreamException,
    KodoResource,
    PtySize,
    ReadyCmd,
    Sandbox,
    SandboxClient,
    Template,
    WatchHandle,
    WriteEntry,
    WriteInfo,
    wait_for_file,
    wait_for_port,
    wait_for_process,
    wait_for_timeout,
    wait_for_url,
)
from .services.sms.sms import Sms
from .services.pili.rtc_server_manager import RtcServer, get_room_token
from .utils import urlsafe_base64_encode, urlsafe_base64_decode, etag, entry, decode_entry, canonical_mime_header_key



# -*- coding: utf-8 -*-
import base64
import hashlib
import json as std_json
import os
import posixpath
import time

from qiniu.compat import bytes as bytes_type
from qiniu.compat import is_py2, str as text_type, urlencode

from .constants import DEFAULT_ENDPOINT, DEFAULT_USER


def normalize_endpoint(endpoint=None):
    endpoint = (
        endpoint or
        os.getenv('QINIU_SANDBOX_API_URL') or
        os.getenv('QINIU_SANDBOX_ENDPOINT') or
        os.getenv('E2B_API_URL') or
        DEFAULT_ENDPOINT
    )
    return endpoint.rstrip('/')


def encode_path(value):
    if is_py2:
        from urllib import quote
        if isinstance(value, text_type):
            value = value.encode('utf-8')
        else:
            value = str(value)
    else:
        from urllib.parse import quote
        value = str(value)
    return quote(value, safe='')


def append_query(path, query=None):
    query = query or {}
    normalized = {}
    for key, value in query.items():
        if value is None:
            continue
        if isinstance(value, (list, tuple)):
            value = ','.join([str(item) for item in value])
        normalized[key] = value
    if not normalized:
        return path
    return path + '?' + urlencode(normalized)


def json_dumps(data):
    return std_json.dumps(data, separators=(',', ':'))


def basic_auth(user=None):
    user = user or DEFAULT_USER
    raw = ('{0}:'.format(user)).encode('utf-8')
    return 'Basic {0}'.format(base64.b64encode(raw).decode('ascii'))


def file_signature(path, operation, user, access_token, expiration):
    if expiration is None:
        expiration = ''
    components = [path, operation, user, access_token, expiration]
    raw = b':'.join([_to_utf8_bytes(component)
                     for component in components])
    return 'v1_{0}'.format(hashlib.sha256(raw).hexdigest())


def _to_utf8_bytes(value):
    if isinstance(value, bytes_type):
        return value
    if isinstance(value, text_type):
        return value.encode('utf-8')
    return str(value).encode('utf-8')


def file_basename(path):
    basename = posixpath.basename(path.rstrip('/')) or 'file'
    if is_py2 and isinstance(basename, text_type):
        return basename.encode('utf-8')
    return basename


def shell_quote(value):
    try:
        from shlex import quote
    except ImportError:
        from pipes import quote
    if is_py2:
        if isinstance(value, text_type):
            value = value.encode('utf-8')
        else:
            value = str(value)
    else:
        value = str(value)
    return quote(value)


def utc_timestamp_after(seconds):
    return int(time.time()) + int(seconds)


def parse_json_response(response):
    if response.content in (None, b'', ''):
        return None
    return response.json()


def get_info_value(info, camel_key, snake_key=None):
    info = info or {}
    if camel_key in info:
        return info.get(camel_key)
    if snake_key and snake_key in info:
        return info.get(snake_key)
    return None



# -*- coding: utf-8 -*-
from qiniu import http
from .config import KIRK_HOST


class QcosClient(object):
    """资源管理客户端

    使用应用密钥生成资源管理客户端，可以进一步：
    1、部署服务和容器，获得信息
    2、创建网络资源，获得信息

    属性：
        auth: 应用密钥对，QiniuMacAuth对象
        host: API host，在『内网模式』下使用时，auth=None，会自动使用 apiproxy 服务，只能管理当前容器所在的应用资源。

    接口：
        list_stacks()
        create_stack(args)
        delete_stack(stack)
        get_stack(stack)
        start_stack(stack)
        stop_stack(stack)

        list_services(stack)
        create_service(stack, args)
        get_service_inspect(stack, service)
        start_service(stack, service)
        stop_service(stack, service)
        update_service(stack, service, args)
        scale_service(stack, service, args)
        delete_service(stack, service)
        create_service_volume(stack, service, volume, args)
        extend_service_volume(stack, service, volume, args)
        delete_service_volume(stack, service, volume)

        list_containers(args)
        get_container_inspect(ip)
        start_container(ip)
        stop_container(ip)
        restart_container(ip)

        list_aps()
        create_ap(args)
        search_ap(mode, query)
        get_ap(apid)
        update_ap(apid, args)
        set_ap_port(apid, port, args)
        delete_ap(apid)
        publish_ap(apid, args)
        unpublish_ap(apid)
        get_ap_port_healthcheck(apid, port)
        set_ap_port_container(apid, port, args)
        disable_ap_port(apid, port)
        enable_ap_port(apid, port)
        get_ap_providers()
        get_web_proxy(backend)
    """

    def __init__(self, auth, host=None):
        self.auth = auth
        if auth is None:
            self.host = KIRK_HOST['APIPROXY']
        else:
            self.host = host

    def list_stacks(self):
        """获得服务组列表

        列出当前应用的所有服务组信息。

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回服务组列表[<stack1>, <stack2>, ...]，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks'.format(self.host)
        return self.__get(url)

    def create_stack(self, args):
        """创建服务组

        创建新一个指定名称的服务组，并创建其下的服务。

        Args:
            - args:  服务组描述，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks'.format(self.host)
        return self.__post(url, args)

    def delete_stack(self, stack):
        """删除服务组

        删除服务组内所有服务并销毁服务组。

        Args:
            - stack:  服务所属的服务组名称

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}'.format(self.host, stack)
        return self.__delete(url)

    def get_stack(self, stack):
        """获取服务组

        查看服务组的属性信息。

        Args:
            - stack:  服务所属的服务组名称

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回stack信息，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}'.format(self.host, stack)
        return self.__get(url)

    def start_stack(self, stack):
        """启动服务组

        启动服务组中的所有停止状态的服务。

        Args:
            - stack:  服务所属的服务组名称

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/start'.format(self.host, stack)
        return self.__post(url)

    def stop_stack(self, stack):
        """停止服务组

        停止服务组中所有运行状态的服务。

        Args:
            - stack:  服务所属的服务组名称

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/stop'.format(self.host, stack)
        return self.__post(url)

    def list_services(self, stack):
        """获得服务列表

        列出指定名称的服务组内所有的服务, 返回一组详细的服务信息。

        Args:
            - stack:  服务所属的服务组名称

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result        成功返回服务信息列表[<ervice1>, <service2>, ...]，失败返回{"error": "<errMsg string>"}
            - ResponseInfo  请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services'.format(self.host, stack)
        return self.__get(url)

    def create_service(self, stack, args):
        """创建服务

        创建一个服务，平台会异步地按模板分配资源并部署所有容器。

        Args:
            - stack:  服务所属的服务组名称
            - args:   服务具体描述请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services'.format(self.host, stack)
        return self.__post(url, args)

    def delete_service(self, stack, service):
        """删除服务

        删除指定名称服务，并自动销毁服务已部署的所有容器和存储卷。

        Args:
            - stack:    服务所属的服务组名称
            - service:  服务名

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services/{2}'.format(self.host, stack, service)
        return self.__delete(url)

    def get_service_inspect(self, stack, service):
        """查看服务

        查看指定名称服务的属性。

        Args:
            - stack:    服务所属的服务组名称
            - service:  服务名

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回服务信息，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services/{2}/inspect'.format(self.host, stack, service)
        return self.__get(url)

    def start_service(self, stack, service):
        """启动服务

        启动指定名称服务的所有容器。

        Args:
            - stack:    服务所属的服务组名称
            - service:  服务名

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services/{2}/start'.format(self.host, stack, service)
        return self.__post(url)

    def stop_service(self, stack, service):
        """停止服务

        停止指定名称服务的所有容器。

        Args:
            - stack:    服务所属的服务组名称
            - service:  服务名

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services/{2}/stop'.format(self.host, stack, service)
        return self.__post(url)

    def update_service(self, stack, service, args):
        """更新服务

        更新指定名称服务的配置如容器镜像等参数，容器被重新部署后生效。
        如果指定manualUpdate参数，则需要额外调用 部署服务 接口并指定参数进行部署；处于人工升级模式的服务禁止执行其他修改操作。
        如果不指定manualUpdate参数，平台会自动完成部署。

        Args:
            - stack:    服务所属的服务组名称
            - service:  服务名
            - args:     服务具体描述请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services/{2}'.format(self.host, stack, service)
        return self.__post(url, args)

    def scale_service(self, stack, service, args):
        """扩容/缩容服务

        更新指定名称服务的配置如容器镜像等参数，容器被重新部署后生效。
        如果指定manualUpdate参数，则需要额外调用 部署服务 接口并指定参数进行部署；处于人工升级模式的服务禁止执行其他修改操作。
        如果不指定manualUpdate参数，平台会自动完成部署。

        Args:
            - stack:    服务所属的服务组名称
            - service:  服务名
            - args:     请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services/{2}/scale'.format(self.host, stack, service)
        return self.__post(url, args)

    def create_service_volume(self, stack, service, args):
        """创建存储卷

        为指定名称的服务增加存储卷资源，并挂载到部署的容器中。

        Args:
            - stack:    服务所属的服务组名称
            - service:  服务名
            - args:     请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services/{2}/volumes'.format(self.host, stack, service)
        return self.__post(url, args)

    def extend_service_volume(self, stack, service, volume, args):
        """扩容存储卷

        为指定名称的服务增加存储卷资源，并挂载到部署的容器中。

        Args:
            - stack:    服务所属的服务组名称
            - service:  服务名
            - volume:   存储卷名
            - args:     请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services/{2}/volumes/{3}/extend'.format(self.host, stack, service, volume)
        return self.__post(url, args)

    def delete_service_volume(self, stack, service, volume):
        """删除存储卷

        从部署的容器中移除挂载，并销毁指定服务下指定名称的存储卷, 并重新启动该容器。

        Args:
            - stack:    服务所属的服务组名称
            - service:  服务名
            - volume:   存储卷名

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/stacks/{1}/services/{2}/volumes/{3}'.format(self.host, stack, service, volume)
        return self.__delete(url)

    def list_containers(self, stack=None, service=None):
        """列出容器列表

        列出应用内所有部署的容器, 返回一组容器IP。

        Args:
            - stack:    要列出容器的服务组名(可不填，表示默认列出所有)
            - service:  要列出容器服务的服务名(可不填，表示默认列出所有)

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回容器的ip数组，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/containers'.format(self.host)
        params = {}
        if stack is not None:
            params['stack'] = stack
        if service is not None:
            params['service'] = service
        return self.__get(url, params or None)

    def get_container_inspect(self, ip):
        """查看容器

        查看指定IP的容器，返回容器属性。

        Args:
            - ip:   容器ip

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回容器的信息，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/containers/{1}/inspect'.format(self.host, ip)
        return self.__get(url)

    def start_container(self, ip):
        """启动容器

        启动指定IP的容器。

        Args:
            - ip:   容器ip

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/containers/{1}/start'.format(self.host, ip)
        return self.__post(url)

    def stop_container(self, ip):
        """停止容器

        停止指定IP的容器。

        Args:
            - ip:   容器ip

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/containers/{1}/stop'.format(self.host, ip)
        return self.__post(url)

    def restart_container(self, ip):
        """重启容器

        重启指定IP的容器。

        Args:
            - ip:   容器ip

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/containers/{1}/restart'.format(self.host, ip)
        return self.__post(url)

    def list_aps(self):
        """列出接入点

        列出当前应用的所有接入点。

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回接入点列表，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps'.format(self.host)
        return self.__get(url)

    def create_ap(self, args):
        """申请接入点

        申请指定配置的接入点资源。

        Args:
            - args:   请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回申请到的接入点信息，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps'.format(self.host)
        return self.__post(url, args)

    def search_ap(self, mode, query):
        """搜索接入点

        查看指定接入点的所有配置信息，包括所有监听端口的配置。

        Args:
            - mode:     搜索模式，可以是domain、ip、host
            - query:    搜索文本

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回搜索结果，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/search?{1}={2}'.format(self.host, mode, query)
        return self.__get(url)

    def get_ap(self, apid):
        """查看接入点

        给出接入点的域名或IP，查看配置信息，包括所有监听端口的配置。

        Args:
            - apid:   接入点ID

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回接入点信息，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/{1}'.format(self.host, apid)
        return self.__get(url)

    def update_ap(self, apid, args):
        """更新接入点

        更新指定接入点的配置，如带宽。

        Args:
            - apid:   接入点ID
            - args:   请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/{1}'.format(self.host, apid)
        return self.__post(url, args)

    def set_ap_port(self, apid, port, args):
        """更新接入点端口配置

        更新接入点指定端口的配置。

        Args:
            - apid: 接入点ID
            - port: 要设置的端口号
            - args: 请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/{1}/{2}'.format(self.host, apid, port)
        return self.__post(url, args)

    def delete_ap(self, apid):
        """释放接入点

        销毁指定接入点资源。

        Args:
            - apid: 接入点ID

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/{1}'.format(self.host, apid)
        return self.__delete(url)

    def publish_ap(self, apid, args):
        """绑定自定义域名

        绑定用户自定义的域名，仅对公网域名模式接入点生效。

        Args:
            - apid: 接入点ID
            - args: 请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/{1}/publish'.format(self.host, apid)
        return self.__post(url, args)

    def unpublish_ap(self, apid, args):
        """解绑自定义域名

        解绑用户自定义的域名，仅对公网域名模式接入点生效。

        Args:
            - apid: 接入点ID
            - args: 请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/{1}/unpublish'.format(self.host, apid)
        return self.__post(url, args)

    def get_ap_port_healthcheck(self, apid, port):
        """查看健康检查结果

        检查接入点的指定端口的后端健康状况。

        Args:
            - apid: 接入点ID
            - port: 要设置的端口号

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回健康状况，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/{1}/{2}/healthcheck'.format(self.host, apid, port)
        return self.__get(url)

    def set_ap_port_container(self, apid, port, args):
        """调整后端实例配置

        调整接入点指定后端实例（容器）的配置，例如临时禁用流量等。

        Args:
            - apid: 接入点ID
            - port: 要设置的端口号

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/{1}/{2}/setcontainer'.format(self.host, apid, port)
        return self.__post(url, args)

    def disable_ap_port(self, apid, port):
        """临时关闭接入点端口

        临时关闭接入点端口，仅对公网域名，公网ip有效。

        Args:
            - apid: 接入点ID
            - port: 要设置的端口号

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/{1}/{2}/disable'.format(self.host, apid, port)
        return self.__post(url)

    def enable_ap_port(self, apid, port):
        """开启接入点端口

        开启临时关闭的接入点端口，仅对公网域名，公网ip有效。

        Args:
            - apid: 接入点ID
            - port: 要设置的端口号

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回空dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/{1}/{2}/enable'.format(self.host, apid, port)
        return self.__post(url)

    def get_ap_providers(self):
        """列出入口提供商

        列出当前支持的入口提供商，仅对申请公网IP模式接入点有效。
        注：公网IP供应商telecom=电信，unicom=联通，mobile=移动。

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回接入商列表，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/aps/providers'.format(self.host)
        return self.__get(url)

    def get_web_proxy(self, backend):
        """获取一次性代理地址

        对内网地址获取一个一次性的外部可访问的代理地址

        Args:
            - backend: 后端地址，如："10.128.0.1:8080"

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回代理地址信息，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/v3/webproxy'.format(self.host)
        return self.__post(url, {'backend': backend})

    def __post(self, url, data=None):
        return http._post_with_qiniu_mac(url, data, self.auth)

    def __get(self, url, params=None):
        return http._get_with_qiniu_mac(url, params, self.auth)

    def __delete(self, url):
        return http._delete_with_qiniu_mac(url, None, self.auth)



# -*- coding: utf-8 -*-


class SandboxError(Exception):
    def __init__(self, message, response=None, data=None):
        super(SandboxError, self).__init__(message)
        self.response = response
        self.data = data
        self.status_code = getattr(response, 'status_code', None)


class TemplateBuildError(SandboxError):
    pass


class CommandExitError(SandboxError):
    def __init__(self, result):
        self.result = result
        super(CommandExitError, self).__init__(
            'Command exited with code {0}'.format(result.exit_code)
        )


class InvalidArgumentException(SandboxError):
    pass


class FileNotFoundException(SandboxError):
    pass


class GitAuthException(SandboxError):
    pass


class GitUpstreamException(SandboxError):
    pass



# -*- coding: utf-8 -*-
from qiniu import http, QiniuMacAuth
from .config import KIRK_HOST
from .qcos_api import QcosClient


class AccountClient(object):
    """客户端入口

    使用账号密钥生成账号客户端，可以进一步：
    1、获取和操作账号数据
    2、获得部署的应用的客户端

    属性：
        auth: 账号管理密钥对，QiniuMacAuth对象
        host: API host，在『内网模式』下使用时，auth=None，会自动使用 apiproxy 服务

    接口：
        get_qcos_client(app_uri)
        create_qcos_client(app_uri)
        get_app_keys(app_uri)
        get_valid_app_auth(app_uri)
        get_account_info()
        get_app_region_products(app_uri)
        get_region_products(region)
        list_regions()
        list_apps()
        create_app(args)
        delete_app(app_uri)

    """

    def __init__(self, auth, host=None):
        self.auth = auth
        self.qcos_clients = {}
        if (auth is None):
            self.host = KIRK_HOST['APPPROXY']
        else:
            self.host = host or KIRK_HOST['APPGLOBAL']
        acc, info = self.get_account_info()
        self.uri = acc.get('name')

    def get_qcos_client(self, app_uri):
        """获得资源管理客户端
        缓存，但不是线程安全的
        """

        client = self.qcos_clients.get(app_uri)
        if (client is None):
            client = self.create_qcos_client(app_uri)
            self.qcos_clients[app_uri] = client

        return client

    def create_qcos_client(self, app_uri):
        """创建资源管理客户端

        """

        if (self.auth is None):
            return QcosClient(None)

        products = self.get_app_region_products(app_uri)
        auth = self.get_valid_app_auth(app_uri)

        if products is None or auth is None:
            return None

        return QcosClient(auth, products.get('api'))

    def get_app_keys(self, app_uri):
        """获得账号下应用的密钥

        列出指定应用的密钥，仅当访问者对指定应用有管理权限时有效：
            用户对创建的应用有管理权限。
            用户对使用的第三方应用没有管理权限，第三方应用的运维方有管理权限。

        Args:
            - app_uri: 应用的完整标识

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回秘钥列表，失败返回None
            - ResponseInfo    请求的Response信息
        """

        url = '{0}/v3/apps/{1}/keys'.format(self.host, app_uri)
        return http._get_with_qiniu_mac(url, None, self.auth)

    def get_valid_app_auth(self, app_uri):
        """获得账号下可用的应用的密钥

        列出指定应用的可用密钥

        Args:
            - app_uri: 应用的完整标识

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回可用秘钥列表，失败返回None
            - ResponseInfo    请求的Response信息
        """

        ret, retInfo = self.get_app_keys(app_uri)

        if ret is None:
            return None

        for k in ret:
            if (k.get('state') == 'enabled'):
                return QiniuMacAuth(k.get('ak'), k.get('sk'))

        return None

    def get_account_info(self):
        """获得当前账号的信息

        查看当前请求方（请求鉴权使用的 AccessKey 的属主）的账号信息。

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回用户信息，失败返回None
            - ResponseInfo    请求的Response信息
        """

        url = '{0}/v3/info'.format(self.host)
        return http._get_with_qiniu_mac(url, None, self.auth)

    def get_app_region_products(self, app_uri):
        """获得指定应用所在区域的产品信息

        Args:
            - app_uri: 应用的完整标识

        Returns:
            返回产品信息列表，若失败则返回None
        """
        apps, retInfo = self.list_apps()
        if apps is None:
            return None

        for app in apps:
            if (app.get('uri') == app_uri):
                return self.get_region_products(app.get('region'))

        return

    def get_region_products(self, region):
        """获得指定区域的产品信息

        Args:
            - region: 区域，如："nq"

        Returns:
            返回该区域的产品信息，若失败则返回None
        """

        regions, retInfo = self.list_regions()
        if regions is None:
            return None

        for r in regions:
            if r.get('name') == region:
                return r.get('products')

    def list_regions(self):
        """获得账号可见的区域的信息

        列出当前用户所有可使用的区域。

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回区域列表，失败返回None
            - ResponseInfo    请求的Response信息
        """

        url = '{0}/v3/regions'.format(self.host)
        return http._get_with_qiniu_mac(url, None, self.auth)

    def list_apps(self):
        """获得当前账号的应用列表

        列出所属应用为当前请求方的应用列表。

        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回应用列表，失败返回None
            - ResponseInfo    请求的Response信息
        """

        url = '{0}/v3/apps'.format(self.host)
        return http._get_with_qiniu_mac(url, None, self.auth)

    def create_app(self, args):
        """创建应用

        在指定区域创建一个新应用，所属应用为当前请求方。

        Args:
            - args: 请求参数(json)，参考 http://kirk-docs.qiniu.com/apidocs/

        Returns:
            - result        成功返回所创建的应用信息，若失败则返回None
            - ResponseInfo  请求的Response信息
        """

        url = '{0}/v3/apps'.format(self.host)
        return http._post_with_qiniu_mac(url, args, self.auth)

    def delete_app(self, app_uri):
        """删除应用

        删除指定标识的应用，当前请求方对该应用应有删除权限。

        Args:
            - app_uri: 应用的完整标识

        Returns:
            - result        成功返回空dict{}，若失败则返回None
            - ResponseInfo  请求的Response信息
        """

        url = '{0}/v3/apps/{1}'.format(self.host, app_uri)
        return http._delete_with_qiniu_mac(url, None, self.auth)



# -*- coding: utf-8 -*-
import time

import requests

from .client import SandboxClient
from .commands import Commands
from .constants import DEFAULT_USER, ENVD_PORT, MCP_PORT
from .errors import SandboxError
from .filesystem import Filesystem
from .git import Git
from .pty import Pty
from .util import (
    append_query,
    file_signature,
    get_info_value,
    utc_timestamp_after,
)


try:
    from time import monotonic as _monotonic_time
except ImportError:
    _monotonic_time = time.time


class _ConnectDescriptor(object):
    def __get__(self, obj, cls):
        if obj is None:
            def class_connect(sandbox_id, client=None, timeout=15, **opts):
                client = client or SandboxClient(**opts)
                info = client.connect_sandbox(sandbox_id, timeout=timeout)
                sandbox = cls(
                    client=client,
                    info=info,
                    sandbox_id=sandbox_id,
                    **opts
                )
                sandbox.refresh_envd_token_if_needed()
                return sandbox
            return class_connect

        def instance_connect(timeout=15):
            info = obj.client.connect_sandbox(
                obj.sandbox_id, timeout=timeout)
            obj.update_info(info)
            obj.refresh_envd_token_if_needed()
            return obj
        return instance_connect


class SandboxPaginator(object):
    def __init__(self, client=None, **opts):
        opts = dict(opts)
        client_opts = {}
        for key in ('endpoint', 'api_url', 'api_key', 'access_token',
                    'mac', 'access_key', 'secret_key', 'session', 'timeout'):
            if key in opts:
                client_opts[key] = opts.pop(key)
        self.client = client or SandboxClient(**client_opts)
        self.next_token = opts.pop('nextToken', None) or opts.pop(
            'next_token', None)
        self.opts = opts
        self._has_next = True

    @property
    def has_next(self):
        return bool(self.next_token) or self._has_next

    hasNext = has_next

    def next_items(self, **opts):
        request_opts = dict(self.opts)
        request_opts.update(opts)
        if self.next_token and request_opts.get('nextToken') is None:
            request_opts['nextToken'] = self.next_token
        data = self.client.list_sandboxes_v2(**request_opts) or {}
        items = data if isinstance(data, list) else (
            data.get('items') or data.get('sandboxes') or []
        )
        self.next_token = None if isinstance(data, list) else (
            data.get('nextToken') or data.get('next_token')
        )
        self._has_next = bool(self.next_token)
        return [Sandbox(client=self.client, info=item) for item in items]

    nextItems = next_items


class Sandbox(object):
    connect = _ConnectDescriptor()

    def __init__(self, client=None, info=None, sandbox_id=None, sandboxID=None,
                 envd_url=None, envdAccessToken=None, **client_opts):
        envd_url = envd_url or client_opts.pop('envdUrl', None)
        self.client = client or SandboxClient(**client_opts)
        self.info = info or {}
        self.sandbox_id = (
            sandbox_id or sandboxID or
            self.info.get('sandboxID') or
            self.info.get('sandboxId') or
            self.info.get('sandbox_id') or
            self.info.get('id')
        )
        self.sandboxID = self.sandbox_id
        self.template_id = (
            self.info.get('templateID') or
            self.info.get('templateId') or
            self.info.get('template_id')
        )
        self.templateID = self.template_id
        self.domain = (
            self.info.get('domain') or
            self.info.get('sandboxDomain') or
            self.info.get('sandbox_domain')
        )
        self.sandbox_domain = self.domain
        self.sandboxDomain = self.domain
        self.envd_version = get_info_value(
            self.info, 'envdVersion', 'envd_version')
        self.envdVersion = self.envd_version
        self.envd_access_token = (
            envdAccessToken or
            get_info_value(self.info, 'envdAccessToken', 'envd_access_token')
        )
        self.envdAccessToken = self.envd_access_token
        self.traffic_access_token = get_info_value(
            self.info, 'trafficAccessToken', 'traffic_access_token'
        )
        self.trafficAccessToken = self.traffic_access_token
        self._envd_url = envd_url
        self.files = Filesystem(self)
        self.filesystem = self.files
        self.commands = Commands(self)
        self.pty = Pty(self)
        self.git = Git(self.commands)

    @classmethod
    def create(cls, template=None, client=None, timeout=None, metadata=None,
               envs=None, secure=True, allow_internet_access=True, mcp=None,
               network=None, lifecycle=None, resources=None, injections=None,
               **opts):
        client_opts = {}
        for key in ('endpoint', 'api_url', 'api_key', 'access_token',
                    'mac', 'access_key', 'secret_key', 'session'):
            if key in opts:
                client_opts[key] = opts.pop(key)
        client = client or SandboxClient(**client_opts)
        info = client.create_sandbox(
            template=template,
            timeout=timeout,
            metadata=metadata,
            envs=envs,
            secure=secure,
            allow_internet_access=allow_internet_access,
            mcp=mcp,
            network=network,
            lifecycle=lifecycle,
            resources=resources,
            injections=injections,
            **opts
        )
        sandbox = cls(client=client, info=info)
        sandbox.refresh_envd_token_if_needed()
        return sandbox

    @classmethod
    def list(cls, client=None, **opts):
        return SandboxPaginator(client=client, **opts)

    def update_info(self, info):
        if not info:
            return self
        self.info = info
        self.sandbox_id = (
            info.get('sandboxID') or info.get('sandboxId') or
            info.get('sandbox_id') or self.sandbox_id
        )
        self.sandboxID = self.sandbox_id
        self.template_id = (
            info.get('templateID') or info.get('templateId') or
            info.get('template_id') or self.template_id
        )
        self.templateID = self.template_id
        self.domain = (
            info.get('domain') or info.get('sandboxDomain') or
            info.get('sandbox_domain') or self.domain
        )
        self.sandbox_domain = self.domain
        self.sandboxDomain = self.domain
        self.envd_access_token = (
            get_info_value(info, 'envdAccessToken', 'envd_access_token') or
            self.envd_access_token
        )
        self.envdAccessToken = self.envd_access_token
        self.traffic_access_token = (
            get_info_value(
                info,
                'trafficAccessToken',
                'traffic_access_token',
            ) or self.traffic_access_token
        )
        self.trafficAccessToken = self.traffic_access_token
        self.envd_version = (
            get_info_value(info, 'envdVersion', 'envd_version') or
            self.envd_version
        )
        self.envdVersion = self.envd_version
        return self

    updateInfo = update_info

    def refresh_envd_token_if_needed(self):
        if self.envd_access_token or not self.sandbox_id:
            return self
        try:
            self.update_info(self.get_info())
        except SandboxError:
            pass
        return self

    refreshEnvdTokenIfNeeded = refresh_envd_token_if_needed

    def kill(self):
        return self.client.delete_sandbox(self.sandbox_id)

    def set_timeout(self, timeout):
        return self.client.update_sandbox_timeout(
            self.sandbox_id, timeout=timeout)

    setTimeout = set_timeout

    def refresh(self, duration=None, **opts):
        if duration is not None:
            opts['duration'] = duration
        return self.client.refresh_sandbox(self.sandbox_id, **opts)

    def pause(self):
        return self.client.pause_sandbox(self.sandbox_id)

    beta_pause = pause
    betaPause = pause

    def resume(self, **opts):
        info = self.client.resume_sandbox(self.sandbox_id, **opts)
        self.update_info(info)
        return self

    def update_network(self, network):
        return self.client.update_sandbox(self.sandbox_id, network=network)

    updateNetwork = update_network

    def get_info(self):
        return self.client.get_sandbox(self.sandbox_id)

    getInfo = get_info

    def get_metrics(self, **opts):
        return self.client.get_sandbox_metrics(self.sandbox_id, **opts)

    getMetrics = get_metrics

    def get_logs(self, **opts):
        return self.client.get_sandbox_logs(self.sandbox_id, **opts)

    getLogs = get_logs

    def get_host(self, port):
        if not self.domain:
            return ''
        return '{0}-{1}.{2}'.format(port, self.sandbox_id, self.domain)

    getHost = get_host

    def envd_url(self):
        if self._envd_url:
            return self._envd_url
        if not self.domain:
            raise SandboxError('Sandbox domain is not available')
        return 'https://{0}'.format(self.get_host(ENVD_PORT))

    envdUrl = envd_url

    def get_mcp_url(self):
        return 'https://{0}/mcp'.format(self.get_host(MCP_PORT))

    getMcpUrl = get_mcp_url

    def get_mcp_token(self):
        return self.traffic_access_token

    getMcpToken = get_mcp_token

    def file_url(
            self,
            path,
            operation,
            user=None,
            signature_expiration=300,
            **opts):
        user = user or DEFAULT_USER
        query = {
            'path': path,
            'username': user,
        }
        if self.envd_access_token:
            expiration = opts.get('signatureExpiration', signature_expiration)
            if expiration is not None and expiration < 1000000000:
                expiration = utc_timestamp_after(expiration)
            query['signature'] = file_signature(
                path,
                operation,
                user,
                self.envd_access_token,
                expiration,
            )
            query['signature_expiration'] = expiration
        return self.envd_url() + append_query('/files', query)

    fileUrl = file_url

    def download_url(self, path, **opts):
        return self.file_url(path, 'read', **opts)

    downloadUrl = download_url
    DownloadURL = download_url

    def upload_url(self, path, **opts):
        return self.file_url(path, 'write', **opts)

    uploadUrl = upload_url
    UploadURL = upload_url

    def wait_for_ready(self, timeout=60, interval=1):
        started = _monotonic_time()
        while True:
            elapsed = _monotonic_time() - started
            remaining = None if timeout is None else max(timeout - elapsed, 0)
            request_timeout = 5
            if remaining is not None:
                request_timeout = max(0.1, min(5, remaining))
            try:
                response = self.client.session.get(
                    self.envd_url() + '/health',
                    timeout=request_timeout,
                )
                if response.status_code >= 200 and response.status_code < 300:
                    return self
            except requests.RequestException:
                pass
            if timeout is not None and elapsed >= timeout:
                raise SandboxError('Sandbox envd did not become ready')
            sleep_time = interval
            if remaining is not None:
                sleep_time = min(interval, remaining)
            if sleep_time > 0:
                time.sleep(sleep_time)

    waitForReady = wait_for_ready

    def is_running(self, request_timeout=None):
        if not self.domain:
            return False
        try:
            response = self.client.session.get(
                self.envd_url() + '/health',
                timeout=request_timeout,
            )
        except requests.RequestException:
            return False
        if response.status_code == 502:
            return False
        if response.status_code >= 200 and response.status_code < 300:
            return True
        return False

    isRunning = is_running



class Endpoint:
    @staticmethod
    def from_host(host):
        """
        Autodetect scheme from host string

        Parameters
        ----------
        host: str

        Returns
        -------
        Endpoint
        """
        if '://' in host:
            scheme, host = host.split('://')
            return Endpoint(host=host, default_scheme=scheme)
        else:
            return Endpoint(host=host)

    def __init__(self, host, default_scheme='https'):
        """
        Parameters
        ----------
        host: str
        default_scheme: str
        """
        self.host = host
        self.default_scheme = default_scheme

    def __str__(self):
        return 'Endpoint(host:\'{0}\',default_scheme:\'{1}\')'.format(
            self.host,
            self.default_scheme
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if not isinstance(other, Endpoint):
            raise TypeError('Cannot compare Endpoint with {0}'.format(type(other)))

        return self.host == other.host and self.default_scheme == other.default_scheme

    def get_value(self, scheme=None):
        """
        Parameters
        ----------
        scheme: str

        Returns
        -------
        str
        """
        scheme = scheme if scheme is not None else self.default_scheme
        return ''.join([scheme, '://', self.host])

    def clone(self):
        """
        Returns
        -------
        Endpoint
        """
        return Endpoint(
            host=self.host,
            default_scheme=self.default_scheme
        )
